@@ -79,7 +79,6 @@ OVN_NATIVE_LB_PROTOCOLS = [constants.PROTOCOL_TCP,
                            constants.PROTOCOL_UDP, ]
 OVN_NATIVE_LB_ALGORITHMS = [constants.LB_ALGORITHM_SOURCE_IP_PORT, ]
 EXCEPTION_MSG = "Exception occurred during %s"
-OVN_EVENT_LOCK_NAME = "neutron_ovn_octavia_event_lock"
 
 
 class IPVersionsMixingNotSupportedError(
@@ -109,8 +108,6 @@ def get_neutron_client():
 
 class LogicalRouterPortEvent(row_event.RowEvent):
 
-    driver = None
-
     def __init__(self, driver):
         table = 'Logical_Router_Port'
         events = (self.ROW_CREATE, self.ROW_DELETE)
@@ -124,7 +121,7 @@ class LogicalRouterPortEvent(row_event.RowEvent):
                   '%(event)s, %(row)s',
                   {'event': event,
                    'row': row})
-        if not self.driver or row.gateway_chassis:
+        if row.gateway_chassis:
             return
         if event == self.ROW_CREATE:
             self.driver.lb_create_lrp_assoc_handler(row)
@@ -134,8 +131,6 @@ class LogicalRouterPortEvent(row_event.RowEvent):
 
 class LogicalSwitchPortUpdateEvent(row_event.RowEvent):
 
-    driver = None
-
     def __init__(self, driver):
         table = 'Logical_Switch_Port'
         events = (self.ROW_UPDATE,)
@@ -145,11 +140,15 @@ class LogicalSwitchPortUpdateEvent(row_event.RowEvent):
         self.driver = driver
 
     def run(self, event, row, old):
+        LOG.debug('LogicalSwitchPortUpdateEvent logged, '
+                  '%(event)s, %(row)s',
+                  {'event': event,
+                   'row': row})
         # Get the neutron:port_name from external_ids and check if
         # it's a vip port or not.
         port_name = row.external_ids.get(
             ovn_const.OVN_PORT_NAME_EXT_ID_KEY, '')
-        if self.driver and port_name.startswith(ovn_const.LB_VIP_PORT_PREFIX):
+        if port_name.startswith(ovn_const.LB_VIP_PORT_PREFIX):
             # Handle port update only for vip ports created by
             # this driver.
             self.driver.vip_port_update_handler(row)
@@ -199,21 +198,19 @@ class OvnNbIdlForLb(ovsdb_monitor.OvnIdl):
 
 class OvnProviderHelper(object):
 
-    ovn_nbdb_api_for_events = None
-    ovn_nb_idl_for_events = None
-    ovn_nbdb_api = None
-
     def __init__(self):
         self.requests = queue.Queue()
         self.helper_thread = threading.Thread(target=self.request_handler)
         self.helper_thread.daemon = True
-        atexit.register(self.shutdown)
         self._octavia_driver_lib = o_driver_lib.DriverLibrary()
         self._check_and_set_ssl_files()
         self._init_lb_actions()
-        self.events = [LogicalRouterPortEvent(self),
-                       LogicalSwitchPortUpdateEvent(self)]
-        self.start()
+
+        # NOTE(mjozefcz): This API is only for handling octavia API requests.
+        self.ovn_nbdb = OvnNbIdlForLb()
+        self.ovn_nbdb_api = self.ovn_nbdb.start()
+
+        self.helper_thread.start()
 
     def _init_lb_actions(self):
         self._lb_request_func_maps = {
@@ -253,8 +250,6 @@ class OvnProviderHelper(object):
     def _check_and_set_ssl_files(self):
         # TODO(reedip): Make ovsdb_monitor's _check_and_set_ssl_files() public
         # This is a copy of ovsdb_monitor._check_and_set_ssl_files
-        if OvnProviderHelper.ovn_nbdb_api:
-            return
         priv_key_file = ovn_conf.get_ovn_nb_private_key()
         cert_file = ovn_conf.get_ovn_nb_certificate()
         ca_cert_file = ovn_conf.get_ovn_nb_ca_cert()
@@ -267,25 +262,10 @@ class OvnProviderHelper(object):
         if ca_cert_file:
             Stream.ssl_set_ca_cert_file(ca_cert_file)
 
-    def start(self):
-        # NOTE(mjozefcz): This API is only for handling octavia API requests.
-        if not OvnProviderHelper.ovn_nbdb_api:
-            OvnProviderHelper.ovn_nbdb_api = OvnNbIdlForLb().start()
-
-        # NOTE(mjozefcz): This API is only for handling OVSDB events!
-        if not OvnProviderHelper.ovn_nbdb_api_for_events:
-            OvnProviderHelper.ovn_nb_idl_for_events = OvnNbIdlForLb(
-                event_lock_name=OVN_EVENT_LOCK_NAME)
-            (OvnProviderHelper.ovn_nb_idl_for_events.notify_handler.
-             watch_events(self.events))
-            OvnProviderHelper.ovn_nbdb_api_for_events = (
-                OvnProviderHelper.ovn_nb_idl_for_events.start())
-        self.helper_thread.start()
-
     def shutdown(self):
         self.requests.put({'type': REQ_TYPE_EXIT})
         self.helper_thread.join()
-        self.ovn_nb_idl_for_events.notify_handler.unwatch_events(self.events)
+        self.ovn_nbdb.stop()
 
     @staticmethod
     def _map_val(row, col, key):
@@ -416,7 +396,7 @@ class OvnProviderHelper(object):
         port_name = vip_lp.external_ids.get(ovn_const.OVN_PORT_NAME_EXT_ID_KEY)
         lb_id = port_name[len(ovn_const.LB_VIP_PORT_PREFIX):]
         try:
-            ovn_lbs = self._find_ovn_lbs(lb_id)
+            ovn_lbs = self._find_ovn_lbs_with_retry(lb_id)
         except idlutils.RowNotFound:
             LOG.debug("Loadbalancer %s not found!", lb_id)
             return
@@ -492,6 +472,14 @@ class OvnProviderHelper(object):
             LOG.error(msg)
             raise driver_exceptions.UpdateStatusError(msg)
 
+    @tenacity.retry(
+        retry=tenacity.retry_if_exception_type(idlutils.RowNotFound),
+        wait=tenacity.wait_exponential(),
+        stop=tenacity.stop_after_delay(10),
+        reraise=True)
+    def _find_ovn_lbs_with_retry(self, lb_id, protocol=None):
+        return self._find_ovn_lbs(lb_id, protocol=protocol)
+
     def _find_ovn_lbs(self, lb_id, protocol=None):
         """Find the Loadbalancers in OVN with the given lb_id as its name
 
@@ -526,12 +514,17 @@ class OvnProviderHelper(object):
         raise idlutils.RowNotFound(table='Load_Balancer',
                                    col='name', match=lb_id)
 
-    def _get_or_create_ovn_lb(self, lb_id, protocol, admin_state_up):
+    def _get_or_create_ovn_lb(
+            self, lb_id, protocol, admin_state_up,
+            lb_algorithm=constants.LB_ALGORITHM_SOURCE_IP_PORT):
         """Find or create ovn lb with given protocol
 
            Find the loadbalancer configured with given protocol or
            create required if not found
         """
+        # TODO(mjozefcz): For now we support only one LB algorithm.
+        # As we may extend that in the future we would need to
+        # look here also for lb_algorithm, along with protocol.
         # Make sure that its lowercase - OVN NBDB stores lowercases
         # for this field.
         protocol = protocol.lower()
@@ -558,6 +551,7 @@ class OvnProviderHelper(object):
             lb_info = {
                 'id': lb_id,
                 'protocol': protocol,
+                constants.LB_ALGORITHM: lb_algorithm,
                 'vip_address': ovn_lbs[0].external_ids.get(
                     ovn_const.LB_EXT_IDS_VIP_KEY),
                 'vip_port_id':
@@ -903,6 +897,14 @@ class OvnProviderHelper(object):
                 return True
         return False
 
+    def _are_selection_fields_supported(self):
+        return self.ovn_nbdb_api.is_col_present(
+            'Load_Balancer', 'selection_fields')
+
+    @staticmethod
+    def _get_selection_keys(lb_algorithm):
+        return ovn_const.LB_SELECTION_FIELDS_MAP[lb_algorithm]
+
     def check_lb_protocol(self, lb_id, listener_protocol):
         ovn_lb = self._find_ovn_lbs(lb_id, protocol=listener_protocol)
         if not ovn_lb:
@@ -946,12 +948,18 @@ class OvnProviderHelper(object):
         lr_ref = loadbalancer.get(ovn_const.LB_EXT_IDS_LR_REF_KEY)
         if lr_ref:
             external_ids[ovn_const.LB_EXT_IDS_LR_REF_KEY] = lr_ref
-
+        # In case we have LB algoritm set
+        lb_algorithm = loadbalancer.get(constants.LB_ALGORITHM)
+        kwargs = {
+            'name': loadbalancer[constants.ID],
+            'protocol': protocol,
+            'external_ids': external_ids}
+        if self._are_selection_fields_supported():
+            kwargs['selection_fields'] = self._get_selection_keys(lb_algorithm)
         try:
             self.ovn_nbdb_api.db_create(
-                'Load_Balancer', name=loadbalancer['id'],
-                protocol=protocol,
-                external_ids=external_ids).execute(check_error=True)
+                'Load_Balancer',
+                **kwargs).execute(check_error=True)
             ovn_lb = self._find_ovn_lbs(
                 loadbalancer['id'],
                 protocol=protocol)
@@ -1278,7 +1286,7 @@ class OvnProviderHelper(object):
                 protocol=listener['protocol'])
         except idlutils.RowNotFound:
             LOG.exception(EXCEPTION_MSG, "update of listener")
-            # LB row not found during updating a listner. That is a problem.
+            # LB row not found during update of a listener. That is a problem.
             status['listeners'][0]['provisioning_status'] = constants.ERROR
             status['loadbalancers'][0]['provisioning_status'] = constants.ERROR
             return status
@@ -1362,7 +1370,8 @@ class OvnProviderHelper(object):
         ovn_lb = self._get_or_create_ovn_lb(
             pool['loadbalancer_id'],
             pool['protocol'],
-            pool['admin_state_up'])
+            pool['admin_state_up'],
+            lb_algorithm=pool[constants.LB_ALGORITHM])
         external_ids = copy.deepcopy(ovn_lb.external_ids)
         pool_key = self._get_pool_key(pool['id'],
                                       is_enabled=pool['admin_state_up'])
@@ -1483,8 +1492,8 @@ class OvnProviderHelper(object):
                 pool['loadbalancer_id'], protocol=pool['protocol'])
         except idlutils.RowNotFound:
             LOG.exception(EXCEPTION_MSG, "update of pool")
-            # LB row not found during updating a listner. That is a problem.
-            status['pool'][0]['provisioning_status'] = constants.ERROR
+            # LB row not found during update of a listener. That is a problem.
+            status['pools'][0]['provisioning_status'] = constants.ERROR
             status['loadbalancers'][0]['provisioning_status'] = constants.ERROR
             return status
 
@@ -1535,9 +1544,9 @@ class OvnProviderHelper(object):
             pool_listeners = self._get_pool_listeners(ovn_lb,
                                                       pool_key)
             listener_status = []
-            for l in pool_listeners:
+            for listener in pool_listeners:
                 listener_status.append(
-                    {'id': l,
+                    {'id': listener,
                      'provisioning_status': constants.ACTIVE})
             status['listeners'] = listener_status
         except Exception:
@@ -1601,9 +1610,9 @@ class OvnProviderHelper(object):
                                    "provisioning_status": constants.ACTIVE}]}
             pool_listeners = self._get_pool_listeners(ovn_lb, pool_key)
             listener_status = []
-            for l in pool_listeners:
+            for listener in pool_listeners:
                 listener_status.append(
-                    {'id': l,
+                    {'id': listener,
                      'provisioning_status': constants.ACTIVE})
             status['listeners'] = listener_status
         except Exception:
@@ -1672,9 +1681,9 @@ class OvnProviderHelper(object):
                                    "provisioning_status": constants.ACTIVE}]}
             pool_listeners = self._get_pool_listeners(ovn_lb, pool_key)
             listener_status = []
-            for l in pool_listeners:
+            for listener in pool_listeners:
                 listener_status.append(
-                    {'id': l,
+                    {'id': listener,
                      'provisioning_status': constants.ACTIVE})
             status['listeners'] = listener_status
         except Exception:
@@ -1731,9 +1740,9 @@ class OvnProviderHelper(object):
 
             pool_listeners = self._get_pool_listeners(ovn_lb, pool_key)
             listener_status = []
-            for l in pool_listeners:
+            for listener in pool_listeners:
                 listener_status.append(
-                    {'id': l,
+                    {'id': listener,
                      'provisioning_status': constants.ACTIVE})
             status['listeners'] = listener_status
         except Exception:
@@ -1786,7 +1795,7 @@ class OvnProviderHelper(object):
         neutron_client = get_neutron_client()
         try:
             return neutron_client.create_port(port)
-        except n_exc.IpAddressAlreadyAllocatedClient:
+        except n_exc.IpAddressAlreadyAllocatedClient as e:
             # Sometimes the VIP is already created (race-conditions)
             # Lets get the it from Neutron API.
             ports = neutron_client.list_ports(
@@ -1795,12 +1804,7 @@ class OvnProviderHelper(object):
             if not ports['ports']:
                 LOG.error('Cannot create/get LoadBalancer VIP port with '
                           'fixed IP: %s', vip_d['vip_address'])
-                status = {'loadbalancers': [{
-                    "id": lb_id,
-                    "provisioning_status": constants.ERROR,
-                    "operating_status": constants.ERROR}]}
-                self._update_status_to_octavia(status)
-                return
+                raise e
             # there should only be one port returned
             port = ports['ports'][0]
             LOG.debug('VIP Port already exists, uuid: %s', port['id'])
@@ -1927,12 +1931,13 @@ class OvnProviderHelper(object):
 
 
 class OvnProviderDriver(driver_base.ProviderDriver):
-    _ovn_helper = None
 
     def __init__(self):
         super(OvnProviderDriver, self).__init__()
-        if not OvnProviderDriver._ovn_helper:
-            OvnProviderDriver._ovn_helper = OvnProviderHelper()
+        self._ovn_helper = OvnProviderHelper()
+
+    def __del__(self):
+        self._ovn_helper.shutdown()
 
     def _check_for_supported_protocols(self, protocol):
         if protocol not in OVN_NATIVE_LB_PROTOCOLS:
@@ -1993,6 +1998,7 @@ class OvnProviderDriver(driver_base.ProviderDriver):
         request_info = {'id': pool.pool_id,
                         'loadbalancer_id': pool.loadbalancer_id,
                         'protocol': pool.protocol,
+                        'lb_algorithm': pool.lb_algorithm,
                         'listener_id': pool.listener_id,
                         'admin_state_up': admin_state_up}
         request = {'type': REQ_TYPE_POOL_CREATE,
@@ -2256,5 +2262,10 @@ class OvnProviderDriver(driver_base.ProviderDriver):
             vip_dict['vip_port_id'] = port['id']
             vip_dict['vip_address'] = port['fixed_ips'][0]['ip_address']
         except Exception as e:
-            raise driver_exceptions.DriverError(e)
+            kwargs = {}
+            if hasattr(e, 'message'):
+                kwargs = {'user_fault_string': e.message,
+                          'operator_fault_string': e.message}
+            raise driver_exceptions.DriverError(
+                **kwargs)
         return vip_dict
