@@ -13,7 +13,9 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
+import atexit
 import copy
+import multiprocessing as mp
 
 import mock
 from neutron.common import utils as n_utils
@@ -29,25 +31,20 @@ from ovsdbapp.schema.ovn_northbound import impl_idl as idl_ovn
 # NOTE(mjozefcz): We need base neutron functionals because we need
 # mechanism driver and l3 plugin.
 from neutron.tests.functional import base
+from ovn_octavia_provider import agent as ovn_agent
 from ovn_octavia_provider.common import constants as ovn_const
 from ovn_octavia_provider import driver as ovn_driver
 
 LR_REF_KEY_HEADER = 'neutron-'
 
 
-class TestOctaviaOvnProviderDriver(
+class TestOvnOctaviaBase(
         base.TestOVNFunctionalBase,
         base.BaseLoggingTestCase):
 
     def setUp(self):
-        super(TestOctaviaOvnProviderDriver, self).setUp()
-        # ovn_driver.OvnProviderHelper.ovn_nbdb_api is a class variable.
-        # Set it to None, so that when a worker starts the 2nd test we don't
-        # use the old object.
+        super(TestOvnOctaviaBase, self).setUp()
         idl_ovn.OvnNbApiIdlImpl.ovsdb_connection = None
-        ovn_driver.OvnProviderHelper.ovn_nbdb_api = None
-        ovn_driver.OvnProviderHelper.ovn_nbdb_api_for_events = None
-        ovn_driver.OvnProviderDriver._ovn_helper = None
         # TODO(mjozefcz): Use octavia listeners to provide needed
         # sockets and modify tests in order to verify if fake
         # listener (status) has received valid value.
@@ -70,7 +67,6 @@ class TestOctaviaOvnProviderDriver(
         self.fake_neutron_client.delete_port.return_value = True
         self._local_net_cache = {}
         self._local_port_cache = {'ports': []}
-        self.addCleanup(self.ovn_driver._ovn_helper.shutdown)
         self.core_plugin = directory.get_plugin()
 
     def _mock_show_subnet(self, subnet_id):
@@ -182,9 +178,13 @@ class TestOctaviaOvnProviderDriver(
                 external_ids[
                     ovn_const.LB_EXT_IDS_LS_REFS_KEY] = jsonutils.loads(
                         ls_refs)
-            lbs.append({'name': lb.name, 'protocol': lb.protocol,
-                        'vips': lb.vips, 'external_ids': external_ids})
-
+            lb_dict = {'name': lb.name, 'protocol': lb.protocol,
+                       'vips': lb.vips, 'external_ids': external_ids}
+            try:
+                lb_dict['selection_fields'] = lb.selection_fields
+            except AttributeError:
+                pass
+            lbs.append(lb_dict)
         return lbs
 
     def _get_loadbalancer_id(self, lb_name):
@@ -440,6 +440,12 @@ class TestOctaviaOvnProviderDriver(
                                          LR_REF_KEY_HEADER + vip_net_id))
 
     def _make_expected_lbs(self, lb_data):
+        def _get_lb_field_by_protocol(protocol, field='external_ids'):
+            "Get needed external_ids and pass by reference"
+            lb = [lb for lb in expected_lbs
+                  if lb.get('protocol') == [protocol]]
+            return lb[0].get(field)
+
         if not lb_data or not lb_data.get('model'):
             return []
 
@@ -455,24 +461,23 @@ class TestOctaviaOvnProviderDriver(
         # Lets fetch list of L4 protocols defined for this LB.
         for p in lb_data['pools']:
             expected_protocols.add(p.protocol.lower())
-        for l in lb_data['listeners']:
-            expected_protocols.add(l.protocol.lower())
+        for listener in lb_data['listeners']:
+            expected_protocols.add(listener.protocol.lower())
         # If there is no protocol lets add default - empty [].
         expected_protocols = list(expected_protocols)
         if len(expected_protocols) == 0:
             expected_protocols.append(None)
 
-        expected_lbs = [{'name': lb_data['model'].loadbalancer_id,
-                         'protocol': [protocol] if protocol else [],
-                         'vips': {},
-                         'external_ids': copy.deepcopy(external_ids)}
-                        for protocol in expected_protocols]
-
-        def _get_lb_field_by_protocol(protocol, field='external_ids'):
-            "Get needed external_ids and pass by reference"
-            lb = [lb for lb in expected_lbs
-                  if lb.get('protocol') == [protocol]]
-            return lb[0].get(field)
+        expected_lbs = []
+        for protocol in expected_protocols:
+            lb = {'name': lb_data['model'].loadbalancer_id,
+                  'protocol': [protocol] if protocol else [],
+                  'vips': {},
+                  'external_ids': copy.deepcopy(external_ids)}
+            if self.ovn_driver._ovn_helper._are_selection_fields_supported():
+                lb['selection_fields'] = ovn_const.LB_SELECTION_FIELDS_MAP[
+                    o_constants.LB_ALGORITHM_SOURCE_IP_PORT]
+            expected_lbs.append(lb)
 
         # For every connected subnet to the LB set the ref
         # counter.
@@ -526,28 +531,28 @@ class TestOctaviaOvnProviderDriver(
             external_ids[pool_key] = p_members
             pool_info[p.pool_id] = p_members
 
-        for l in lb_data['listeners']:
+        for listener in lb_data['listeners']:
             expected_vips = _get_lb_field_by_protocol(
-                l.protocol.lower(),
+                listener.protocol.lower(),
                 field='vips')
             external_ids = _get_lb_field_by_protocol(
-                l.protocol.lower(),
+                listener.protocol.lower(),
                 field='external_ids')
-            listener_k = 'listener_' + str(l.listener_id)
-            if lb_data['model'].admin_state_up and l.admin_state_up:
+            listener_k = 'listener_' + str(listener.listener_id)
+            if lb_data['model'].admin_state_up and listener.admin_state_up:
                 vip_k = lb_data['model'].vip_address + ":" + str(
-                    l.protocol_port)
-                if not isinstance(l.default_pool_id,
+                    listener.protocol_port)
+                if not isinstance(listener.default_pool_id,
                                   octavia_data_model.UnsetType) and pool_info[
-                                      l.default_pool_id]:
+                                      listener.default_pool_id]:
                     expected_vips[vip_k] = self._extract_member_info(
-                        pool_info[l.default_pool_id])
+                        pool_info[listener.default_pool_id])
             else:
                 listener_k += ':D'
-            external_ids[listener_k] = str(l.protocol_port) + ":"
-            if not isinstance(l.default_pool_id,
+            external_ids[listener_k] = str(listener.protocol_port) + ":"
+            if not isinstance(listener.default_pool_id,
                               octavia_data_model.UnsetType):
-                external_ids[listener_k] += 'pool_' + (l.default_pool_id)
+                external_ids[listener_k] += 'pool_' + listener.default_pool_id
             elif lb_data.get('pools', []):
                 external_ids[listener_k] += 'pool_' + lb_data[
                     'pools'][0].pool_id
@@ -601,8 +606,8 @@ class TestOctaviaOvnProviderDriver(
 
         pool_listeners = self._get_pool_listeners(lb_data, m_pool.pool_id)
         expected_listener_status = [
-            {'id': l.listener_id, 'provisioning_status': 'ACTIVE'}
-            for l in pool_listeners]
+            {'id': listener.listener_id, 'provisioning_status': 'ACTIVE'}
+            for listener in pool_listeners]
         self.ovn_driver.pool_update(m_pool, m_pool)
         expected_status = {
             'pools': [{'id': m_pool.pool_id,
@@ -669,16 +674,16 @@ class TestOctaviaOvnProviderDriver(
                 return p
 
     def _get_listener_from_lb_data(self, lb_data, protocol, protocol_port):
-        for l in lb_data['listeners']:
-            if (l.protocol_port == protocol_port and
-                    l.protocol == protocol):
-                return l
+        for listener in lb_data['listeners']:
+            if (listener.protocol_port == protocol_port and
+                    listener.protocol == protocol):
+                return listener
 
     def _get_pool_listeners(self, lb_data, pool_id):
         listeners = []
-        for l in lb_data['listeners']:
-            if l.default_pool_id == pool_id:
-                listeners.append(l)
+        for listener in lb_data['listeners']:
+            if listener.default_pool_id == pool_id:
+                listeners.append(listener)
 
         return listeners
 
@@ -697,8 +702,8 @@ class TestOctaviaOvnProviderDriver(
         self._update_ls_refs(lb_data, network_id)
         pool_listeners = self._get_pool_listeners(lb_data, pool_id)
         expected_listener_status = [
-            {'id': l.listener_id, 'provisioning_status': 'ACTIVE'}
-            for l in pool_listeners]
+            {'id': listener.listener_id, 'provisioning_status': 'ACTIVE'}
+            for listener in pool_listeners]
 
         expected_status = {
             'pools': [pool_status],
@@ -873,6 +878,10 @@ class TestOctaviaOvnProviderDriver(
 
         self._wait_for_status_and_validate(lb_data, [expected_status])
 
+
+class TestOvnOctaviaProviderDriver(
+        TestOvnOctaviaBase):
+
     def test_loadbalancer(self):
         lb_data = self._create_load_balancer_and_validate(
             {'vip_network': 'vip_network',
@@ -1128,6 +1137,66 @@ class TestOctaviaOvnProviderDriver(
         self._create_listener_and_validate(lb_data)
         self._delete_load_balancer_and_validate(lb_data)
 
+    def test_lb_listener_pool_workflow(self):
+        lb_data = self._create_load_balancer_and_validate(
+            {'vip_network': 'vip_network',
+             'cidr': '10.0.0.0/24'})
+        self._create_listener_and_validate(lb_data)
+        self._create_pool_and_validate(
+            lb_data, "p1", listener_id=lb_data['listeners'][0].listener_id)
+        self._delete_pool_and_validate(
+            lb_data, "p1", listener_id=lb_data['listeners'][0].listener_id)
+        self._delete_listener_and_validate(lb_data)
+        self._delete_load_balancer_and_validate(lb_data)
+
+    def test_lb_member_batch_update(self):
+        # Create a LoadBalancer
+        lb_data = self._create_load_balancer_and_validate(
+            {'vip_network': 'vip_network',
+             'cidr': '10.0.0.0/24'})
+        # Create a pool
+        self._create_pool_and_validate(lb_data, "p1")
+        pool_id = lb_data['pools'][0].pool_id
+        # Create Member-1 and associate it with lb_data
+        self._create_member_and_validate(
+            lb_data, pool_id, lb_data['vip_net_info'][1],
+            lb_data['vip_net_info'][0], '10.0.0.10')
+        # Create Member-2
+        m_member = self._create_member_model(pool_id,
+                                             lb_data['vip_net_info'][1],
+                                             '10.0.0.12')
+        # Update ovn's Logical switch reference
+        self._update_ls_refs(lb_data, lb_data['vip_net_info'][0])
+        lb_data['pools'][0].members.append(m_member)
+        # Add a new member to the LB
+        members = [m_member] + [lb_data['pools'][0].members[0]]
+        self._update_members_in_batch_and_validate(lb_data, pool_id, members)
+        # Deleting one member, while keeping the other member available
+        self._update_members_in_batch_and_validate(lb_data, pool_id,
+                                                   [m_member])
+        self._delete_load_balancer_and_validate(lb_data)
+
+
+class TestOvnOctaviaProviderAgent(TestOvnOctaviaBase):
+
+    def setUp(self):
+        super(TestOvnOctaviaProviderAgent, self).setUp()
+        self._initialize_ovn_da()
+
+    def _initialize_ovn_da(self):
+        # NOTE(mjozefcz): In theory this is separate process
+        # with IDL running, but to make it easier for now
+        # we can initialize this IDL here instead spawning
+        # another process.
+        da_helper = ovn_driver.OvnProviderHelper()
+        events = [ovn_driver.LogicalRouterPortEvent(da_helper),
+                  ovn_driver.LogicalSwitchPortUpdateEvent(da_helper)]
+        ovn_nb_idl_for_events = ovn_driver.OvnNbIdlForLb(
+            event_lock_name='func_test')
+        ovn_nb_idl_for_events.notify_handler.watch_events(events)
+        ovn_nb_idl_for_events.start()
+        atexit.register(da_helper.shutdown)
+
     def _test_lrp_event_handler(self, cascade=False):
         # Create Network N1 on router R1 and LBA on N1
         lba_data = self._create_load_balancer_and_validate(
@@ -1243,45 +1312,6 @@ class TestOctaviaOvnProviderDriver(
                 LR_REF_KEY_HEADER + provider_net['network']['id']),
             timeout=10)
 
-    def test_lb_listener_pool_workflow(self):
-        lb_data = self._create_load_balancer_and_validate(
-            {'vip_network': 'vip_network',
-             'cidr': '10.0.0.0/24'})
-        self._create_listener_and_validate(lb_data)
-        self._create_pool_and_validate(
-            lb_data, "p1", listener_id=lb_data['listeners'][0].listener_id)
-        self._delete_pool_and_validate(
-            lb_data, "p1", listener_id=lb_data['listeners'][0].listener_id)
-        self._delete_listener_and_validate(lb_data)
-        self._delete_load_balancer_and_validate(lb_data)
-
-    def test_lb_member_batch_update(self):
-        # Create a LoadBalancer
-        lb_data = self._create_load_balancer_and_validate(
-            {'vip_network': 'vip_network',
-             'cidr': '10.0.0.0/24'})
-        # Create a pool
-        self._create_pool_and_validate(lb_data, "p1")
-        pool_id = lb_data['pools'][0].pool_id
-        # Create Member-1 and associate it with lb_data
-        self._create_member_and_validate(
-            lb_data, pool_id, lb_data['vip_net_info'][1],
-            lb_data['vip_net_info'][0], '10.0.0.10')
-        # Create Member-2
-        m_member = self._create_member_model(pool_id,
-                                             lb_data['vip_net_info'][1],
-                                             '10.0.0.12')
-        # Update ovn's Logical switch reference
-        self._update_ls_refs(lb_data, lb_data['vip_net_info'][0])
-        lb_data['pools'][0].members.append(m_member)
-        # Add a new member to the LB
-        members = [m_member] + [lb_data['pools'][0].members[0]]
-        self._update_members_in_batch_and_validate(lb_data, pool_id, members)
-        # Deleting one member, while keeping the other member available
-        self._update_members_in_batch_and_validate(lb_data, pool_id,
-                                                   [m_member])
-        self._delete_load_balancer_and_validate(lb_data)
-
     def test_fip_on_lb_vip(self):
         """This test checks if FIP on LB VIP is configured.
 
@@ -1351,3 +1381,13 @@ class TestOctaviaOvnProviderDriver(
             elif ls.name == provider_net:
                 # Make sure that LB1 is not added to provider net - e1 LS
                 self.assertListEqual([], ls.load_balancer)
+
+    def test_agent_exit(self):
+        exit_event = mp.Event()
+        agent = mp.Process(target=ovn_agent.OvnProviderAgent,
+                           args=[exit_event])
+        agent.start()
+        self.assertTrue(agent.is_alive())
+        exit_event.set()
+        agent.join()
+        self.assertFalse(agent.is_alive())
